# Check to see whether doing behavioral or structural simulation
# based upon the current directory
set current_dir [pwd]
if { [ regexp {simulation/behavioral$} $current_dir ] } {
    echo "Performing a behavioral simulation."
    set BEHAVIORAL_SIM 1
} elseif { [ regexp {simulation/structural$} $current_dir ] } {
    echo "Performing a structural simulation."
    set BEHAVIORAL_SIM 0
} else {
    echo "Current directory is not simulation/behavioral nor simulation/structural"
    error "Testbench supports only behavioral and structural simulation...quitting"
}

# Set PPC_MODEL_BFM to 0 if using SWIFT MODEL of PPC405 (default)
set PPC_MODEL_BFM 0

# echo simulation type
if {$PPC_MODEL_BFM == 1} {
    echo "Using Bus-Functional Model (BFM) of PPC405"
} else {
    echo "Using SWIFT Model of PPC405"
}

# Alter system_init.v file to use the testbench hierarchy
if { [file readable "system_init.v"] } {
    echo
    echo "Modifying system_init.v to use the top-level instance name"
    echo
    exec sh ../../data/init_patch.sh
}

# copy IDCT ROM initialization files
cp ../../data/dct2d_top_final3_2d_idct1_ROM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_ROM1.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem1_demux_mux1_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem1_demux_mux2_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem1_demux_mux3_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem1_demux_mux4_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem2_demux_mux1_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem2_demux_mux2_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem2_demux_mux3_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem2_demux_mux4_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem3_demux_mux1_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem3_demux_mux2_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem3_demux_mux3_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem3_demux_mux4_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem4_demux_mux1_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem4_demux_mux2_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem4_demux_mux3_Dual_Port_RAM.mif .
cp ../../data/dct2d_top_final3_2d_idct1_Subsystem4_Subsystem4_demux_mux4_Dual_Port_RAM.mif .
cp ../../data/dual_port_block_memory_virtex4_6_1_91317ac2c213573c.mif .
cp ../../data/single_port_block_memory_virtex4_6_1_08b712684228d44d.mif .
cp ../../data/single_port_block_memory_virtex4_6_1_afb5bdd460500d52.mif .

# copy a modified system.v that has simulation delays on APUFCM* signals
# cp ../../data/system.v .

# compile design files
do system.do

# compile testbench models
set testbench_defines "+define+PPC_MODEL_SWIFT"
set testbench_defines "$testbench_defines+BEHAVIORAL_SIM"
vlog $testbench_defines -vlog95compat -incr -work work ../../data/testbench.v

# compile in SWIFT files
vsim -t 1ps -Lf XilinxCoreLib_ver -Lf simprims_ver -L unisims_ver -c work.testbench work.system work.glbl

# start wave viewer
do ../../data/wave.do


