//-------------------------------------------------------------------------
// TestBench - Module
//-------------------------------------------------------------------------
//
//     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
//     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
//     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
//     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
//     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
//     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
//     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
//     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
//     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
//     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
//     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//     FOR A PARTICULAR PURPOSE.
//     
//     (c) Copyright 2005 Xilinx, Inc.
//     All rights reserved.
//
//-------------------------------------------------------------------------
// Filename: testbench.v
// 
// Description:    
//   Test Bench Top Level wrapper
//
// Design Notes:
//   Contains behavioral code to connect the device under test
//   (Virtex-II Pro FPGA) to external devices for testing.
//
//-------------------------------------------------------------------------
// Author:      
// History:
//
//-------------------------------------------------------------------------
// Naming Conventions:
//      active low signals:                   "*_n"
//      clock signals:                        "clk", "clk_div#", "clk_#x" 
//      reset signals:                        "rst", "rst_n" 
//      generics/parameters:                  "C_*" 
//      user defined types:                   "*_TYPE" 
//      state machine next state:             "*_ns" 
//      state machine current state:          "*_cs" 
//      combinatorial signals:                "*_com" 
//      pipelined or register delay signals:  "*_d#" 
//      counter signals:                      "*cnt*"
//      clock enable signals:                 "*_ce" 
//      internal version of output port       "*_i"
//      device pins:                          "*_pin" 
//      ports:                                - Names begin with Uppercase 
//      processes:                            "*_PROCESS" 
//      component instantiations:             "<ENTITY_>I_<#|FUNC>
//-------------------------------------------------------------------------

`timescale  1 ns / 1 ps

module testbench ();
   
  ///////////////////////////////////////////////////////////////////////////
  // Signal Declarations
  ///////////////////////////////////////////////////////////////////////////
  
  // Internal Signals
  reg         clk;
  reg 	      external_rst_n;
  reg         gsr_rst;
  // UART 1
  reg         uart_RX;
  wire        uart_TX;
  // DDR SDRAM
  wire        ddr_clk;
  wire        ddr_clk_fb;

  // Display Testbench Messages on/off Flags
  wire          reg_MsgOn;
   
  // Parameters for Simulation
`include "../../data/sim_params.v"
   
  ///////////////////////////////////////////////////////////////////////////
  // FPGA Instantiation as Device Under Test
  ///////////////////////////////////////////////////////////////////////////

  // BRAM initialization
  system_conf system_conf_0();
   
  system system_0
    (
     .uart_RX(uart_RX),
     .uart_TX(uart_TX),
     // DDR SDRAM
     .ddr_clk(ddr_clk),
     .ddr_clk_fb(ddr_clk_fb),
     .sys_clk_in(clk),
     .sys_rst_in(external_rst_n)
     );
  
  ///////////////////////////////////////////////////////////////////////////
  // Generate Clock and Resets for Testbench
  ///////////////////////////////////////////////////////////////////////////

  // Initialize clock and resets 
  initial begin
     clk = 0;
     external_rst_n = 1; 
     gsr_rst = 1;
     uart_RX = 1'b0;
     #(`SIM_GSR_DELAY) gsr_rst = 0;
     $display ("\nGSR Deasserted at time = %t", $realtime);
  end
   
  assign glbl.GSR = gsr_rst; // assert reset to Xilinx sim primitives
  assign ddr_clk_fb = ddr_clk; // feed DDR clock back to FPGA
   
  always #(`SIM_REF_CLOCK_PERIOD*0.5) clk = ~clk; // Main clock
   
  // stop simulation if max sim time (specified in sim_params.v) is exceeded
  initial begin
    if (`SIM_MAX_SIM_TIME > 0) begin // Set max time to 0 to bypass
      #(`SIM_MAX_SIM_TIME);
        $stop;
    end
  end

  ///////////////////////////////////////////////////////////////////////////
  // Swift Model Simulation Related Commands
  ///////////////////////////////////////////////////////////////////////////

`ifdef PPC_MODEL_SWIFT
`ifdef STRUCTURAL_SIM
   initial $lm_monitor_enable(`PPC405_PATH.IPPC405_SWIFT.I1);
`endif
   // VIEW_REG: display register updates in CPU swift model
   `include "../../data/view_reg.v"
`endif
   
endmodule // testbench
