/////////////////////////////////////////////////////////////////////////
//
//     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
//     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
//     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
//     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
//     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
//     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
//     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
//     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
//     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
//     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
//     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
//     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
//     FOR A PARTICULAR PURPOSE.
//     
//     (c) Copyright 2005 Xilinx, Inc.
//     All rights reserved.
//
/////////////////////////////////////////////////////////////////////////

/********************************************************************
 * FCM module to test APU-FCM execution of FCM Load and and FCM Store 
 * instructions. This modules contains an interface to the APU and 
 * has two 32-entry register files with 32-bit registers. One regfile
 * - called the Load regfile is used for loads and the other, for
 * stores.
 * The second part of this module interfaces with an IDCT module to
 * implement transfers of memory data for IDCT operations.
 * 
 * Author: APD Systems Engineering
 * Date: Aug 17, 2004
 * History:
 * 
 *******************************************************************/
`timescale 1 ns / 1 ps

module fcm 
  (
   // outputs 
   // to APU
   FCMAPUINSTRACK,
   FCMAPURESULT,
   FCMAPUDONE,
   FCMAPUSLEEPNOTREADY,
   FCMAPUDECODEBUSY,
   FCMAPUDCDGPRWRITE,
   FCMAPUDCDRAEN,
   FCMAPUDCDRBEN,
   FCMAPUDCDPRIVOP,
   FCMAPUDCDFORCEALIGN,
   FCMAPUDCDXEROVEN,
   FCMAPUDCDXERCAEN,
   FCMAPUDCDCREN,
   FCMAPUEXECRFIELD,
   FCMAPUDCDLOAD,
   FCMAPUDCDSTORE,
   FCMAPUDCDUPDATE,
   FCMAPUDCDLDSTBYTE,
   FCMAPUDCDLDSTHW,
   FCMAPUDCDLDSTWD,
   FCMAPUDCDLDSTDW,
   FCMAPUDCDLDSTQW,
   FCMAPUDCDTRAPLE,
   FCMAPUDCDTRAPBE,
   FCMAPUDCDFORCEBESTEERING,
   FCMAPUDCDFPUOP,
   FCMAPUEXEBLOCKINGMCO,
   FCMAPUEXENONBLOCKINGMCO,
   FCMAPULOADWAIT,
   FCMAPURESULTVALID,
   FCMAPUXEROV,
   FCMAPUXERCA,
   FCMAPUCR,
   FCMAPUEXCEPTION,

   // inputs 
   // from APU
   APUFCMINSTRUCTION,
   APUFCMINSTRVALID,
   APUFCMRADATA,
   APUFCMRBDATA,
   APUFCMOPERANDVALID,
   APUFCMFLUSH,
   APUFCMWRITEBACKOK,
   APUFCMLOADDATA,
   APUFCMLOADDVALID,
   APUFCMLOADBYTEEN,
   APUFCMENDIAN,
   APUFCMXERCA,
   APUFCMDECODED,
   APUFCMDECUDI,
   APUFCMDECUDIVALID,

   // clock and reset
   clock,
   reset
   
   );

   // state mnemonics
   parameter // load/store states
	     STATE_IDLE  = 2'b00, // idle state
	     STATE_LOAD  = 2'b01, // valid load instruction detected
	     STATE_STORE = 2'b10, // valid store instruction detected
	     STATE_WAITIDCT = 2'b11, // wait for idct result before returning data
	     // idct states
	     IDCT_STATE_IDLE      = 2'b00, // idct idle state
	     IDCT_STATE_SENDDATA  = 2'b01, // send data to idct
	     IDCT_STATE_WAITIDCT  = 2'b10, // wait for data from idct
   	     IDCT_STATE_RECVDATA  = 2'b11, // store idct results
	     NUM_PIXELROW = 7; // 0-7 = eight rows = four words
	     
   // port declarations
   // outputs
   // to APU
   output        FCMAPUINSTRACK;
   output [0:31] FCMAPURESULT;
   output 	 FCMAPUDONE;
   output 	 FCMAPUSLEEPNOTREADY;
   output 	 FCMAPUDECODEBUSY;
   output 	 FCMAPUDCDGPRWRITE;
   output 	 FCMAPUDCDRAEN;
   output 	 FCMAPUDCDRBEN;
   output 	 FCMAPUDCDPRIVOP;
   output 	 FCMAPUDCDFORCEALIGN;
   output 	 FCMAPUDCDXEROVEN;
   output 	 FCMAPUDCDXERCAEN;
   output 	 FCMAPUDCDCREN;
   output [0:2]  FCMAPUEXECRFIELD;
   output 	 FCMAPUDCDLOAD;
   output 	 FCMAPUDCDSTORE;
   output 	 FCMAPUDCDUPDATE;
   output 	 FCMAPUDCDLDSTBYTE;
   output 	 FCMAPUDCDLDSTHW;
   output 	 FCMAPUDCDLDSTWD;
   output 	 FCMAPUDCDLDSTDW;
   output 	 FCMAPUDCDLDSTQW;
   output 	 FCMAPUDCDTRAPLE;
   output 	 FCMAPUDCDTRAPBE;
   output 	 FCMAPUDCDFORCEBESTEERING;
   output 	 FCMAPUDCDFPUOP;
   output 	 FCMAPUEXEBLOCKINGMCO;
   output 	 FCMAPUEXENONBLOCKINGMCO;
   output 	 FCMAPULOADWAIT;
   output 	 FCMAPURESULTVALID;
   output 	 FCMAPUXEROV;
   output 	 FCMAPUXERCA;
   output [0:3]  FCMAPUCR;
   output 	 FCMAPUEXCEPTION;
	
   // inputs
   // to APU
   input [0:31]  APUFCMINSTRUCTION;
   input 	 APUFCMINSTRVALID;
   input [0:31]	 APUFCMRADATA;
   input [0:31]	 APUFCMRBDATA;
   input 	 APUFCMOPERANDVALID;
   input 	 APUFCMFLUSH;
   input 	 APUFCMWRITEBACKOK;
   input [0:31]  APUFCMLOADDATA;
   input 	 APUFCMLOADDVALID;
   input [0:3] 	 APUFCMLOADBYTEEN;
   input 	 APUFCMENDIAN;
   input 	 APUFCMXERCA;
   input 	 APUFCMDECODED;
   input [0:2] 	 APUFCMDECUDI;
   input 	 APUFCMDECUDIVALID;
   // clock and reset
   input 	 clock;
   input 	 reset;
   
   // internal signals
   // PPC instruction-related
   reg [0:5] 	 reg_RT; // target register of PPC instruction
   reg [0:5] 	 reg_RA; // base register of PPC instruction
   reg [0:5] 	 reg_RB; // offset register of PPC instruction
   wire 	 instrreg_we; // write enable
   // loads and stores
   wire 	 store_or_loadn; // 1=store, 0=load
   reg 		 store_or_loadn_reg; // stores the store_or_loadn signal
   wire 	 ldst_update; // 1=update RA
   wire [0:1] 	 ldst_size; // number of words to transfer
   reg [0:1] 	 ldst_size_reg; // stores number of words to transfer
   reg [0:1] 	 ldst_size_counter; // counter for number of words
   wire 	 ldst_size_counter_we; // write enable   
   wire 	 ldst_valid; // 1=valid load/store instruction   
   // register file
   reg [0:31] 	 regfile [0:3]; // 32-bit entry register file with 4 registers
   wire 	 regfile_we; // write enable
   wire [0:1] 	 regfile_waddr; // write address
   wire [0:31] 	 regfile_wdata; // write data
   reg [0:4] 	 regfile_raddr; // register to store read address
   wire 	 regfile_raddr_we; // write enable 	 
   wire [0:31] 	 regfile_rdata; // read data
   wire 	 regfile_readvalid; // read data is valid
   // state registers
   reg [0:1] 	 curr_state; // current state
   reg [0:1] 	 next_state; // next state

   // internal variables
   // outputs to IDCT module
   wire [0:127]  fcmidctdata; // data to give to idct
   wire 	 fcmidctready; // 1=tells idct to accept input data
   // inputs from IDCT module
   wire [0:127]  idctfcmdata; // processed data from idct
   wire 	 idctfcmready; // signal from idct that results are ready
   // register file for IDCT values and FCM stores
   reg [0:31] 	 regfile_store[0:15]; // 32-bit entry register file with 16 registers
   wire 	 regfile_store_we; // write enable
   wire [0:3] 	 regfile_store_base; // block (1 block == 4 words) addr
   // idct pixel blocks
   reg [0:2] 	 pixelrow_counter; // 1 row = 8 pixels = 4 words
   wire 	 pixelrow_counter_we; // write enable
   // state registers
   reg [0:1] idct_curr_state;
   reg [0:1] idct_next_state;
   
   /****************************** sequential blocks **************************/
   // PPC instruction-related registers
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  begin
	     reg_RT <= 5'b0;
	     reg_RA <= 5'b0;
	     reg_RB <= 5'b0;
	     store_or_loadn_reg <= 1'b0;
	     ldst_size_reg <= 2'b0;
	  end
	else
	  if (instrreg_we) // capture instruction information
	    begin
	       reg_RT <= APUFCMINSTRUCTION[6:10];
	       reg_RA <= APUFCMINSTRUCTION[11:15];
	       reg_RB <= APUFCMINSTRUCTION[16:20];
	       store_or_loadn_reg <= store_or_loadn;
	       ldst_size_reg <= ldst_size;
	    end
     end

   // read address of store register file
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  regfile_raddr <= 5'h0;
	else
	  if (instrreg_we)
	    regfile_raddr <= APUFCMINSTRUCTION[6:10];
	  else
	    if (regfile_raddr_we)
	      regfile_raddr <= regfile_raddr + 1;
     end // always @ (posedge clock or posedge reset)
   
   // load / store counter for number of words
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  ldst_size_counter <= 2'b0;
	else
	  if (instrreg_we)
	    ldst_size_counter <= 2'b0;
	  else
	    if (ldst_size_counter_we)
	      ldst_size_counter <= ldst_size_counter + 1;
     end // always @ (posedge clock or posedge reset)

   // register file for loads
   always @(posedge clock or posedge reset)
     begin
	if (reset) // set all register values to zero
	  begin
	     regfile[0]  <= 32'b0; regfile[1]  <= 32'b0; 
	     regfile[2]  <= 32'b0; regfile[3]  <= 32'b0; 
	  end // if (reset)
	else
	  if (regfile_we) // write data to specified register when write enabled
	    regfile[regfile_waddr] <= regfile_wdata;
     end

   // state machine
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  begin
	     curr_state <= STATE_IDLE;
	  end
	else
	  curr_state <= next_state;
     end

   /**************************** combinational blocks *************************/
   // decoder
   decode_ldst decode_ldst_0 
     ( // outputs
       .update(ldst_update),
       .size(ldst_size),
       .store_or_loadn(store_or_loadn),
       .valid_ldst(ldst_valid),
       // inputs
       .APUFCMINSTRUCTION(APUFCMINSTRUCTION)
       );
   
   // state machine logic
   always @(curr_state or store_or_loadn or store_or_loadn_reg or 
	    ldst_size_counter or ldst_size_reg or
	    idct_curr_state or idctfcmready or
	    APUFCMINSTRVALID or 
	    APUFCMLOADDVALID or
	    APUFCMDECODED)
     begin
	case (curr_state)

	  // wait for valid instruction
	  STATE_IDLE:
	    if (APUFCMINSTRVALID &
		APUFCMDECODED) // valid instruction from APU
	      if (store_or_loadn) // store instruction
		if (idct_curr_state == IDCT_STATE_WAITIDCT) // waiting for idct
		  next_state = STATE_WAITIDCT;
	  	else
		  next_state = STATE_STORE;
	      else // load instruction
		if (APUFCMLOADDVALID) // load data arrived at the same time
		  if (ldst_size_counter < ldst_size_reg)
		    next_state = STATE_LOAD;
		  else
		    next_state = STATE_IDLE;
		else
		  next_state = STATE_LOAD;
	    else
	      next_state = STATE_IDLE;

	  // seen a valid load instruction, wait for valid data
	  STATE_LOAD:
	    // keep track of how many words to access
	    if (ldst_size_counter < ldst_size_reg)
	      next_state = STATE_LOAD;
	    else
	      if (APUFCMLOADDVALID)
		next_state = STATE_IDLE;
	      else
		next_state = STATE_LOAD;

	  // got a store instruction but IDCT is not ready
	  STATE_WAITIDCT:
	    // idct result is valid
	    if (idctfcmready)
	      next_state = STATE_STORE;
	    else
	      next_state = STATE_WAITIDCT;
	  
	  // seen a valid store instruction, output data
	  STATE_STORE:
	    // keep track of how many words to access
	    if (ldst_size_counter < ldst_size_reg)
	      next_state = STATE_STORE;
	    else
	      next_state = STATE_IDLE;
	  
	  default:
	    next_state = STATE_IDLE;
	  
	endcase // case(curr_state)
     end // always @ (APUFCMINSTRUCTION or APUFCMINSTRVALID or APUFCMRADATA or..

   // internal signal assignments
   assign instrreg_we = (curr_state == STATE_IDLE);
   // data might arrive at the same time as the instruction
   assign regfile_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID) |
			 ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID) );
   // reg_RT is the target register value when data comes after the instruction
   assign regfile_waddr = (reg_RT + ldst_size_counter);
   // write data from memory to register file
   assign regfile_wdata = APUFCMLOADDATA;
   // address of register file to read data from for a store operation
   assign regfile_raddr_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID ) |
			       (curr_state == STATE_STORE) |
			       ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID ) );
   // data read for a store operation
   assign regfile_rdata = regfile_store[regfile_raddr];
   assign regfile_readvalid = 1'b1; // always set to one
   // update counter for number of data words transferred
   assign ldst_size_counter_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID ) |
				   (curr_state == STATE_STORE) |
				   ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID ) );
   
   // output assignments
   assign FCMAPUDONE = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID &
			   (ldst_size_counter == ldst_size_reg) ) |
			 ( (curr_state == STATE_STORE) & 
			   (ldst_size_counter == ldst_size_reg) ) |
			 ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID &
			   (ldst_size_counter == ldst_size_reg) ) ) ? 1'b1 : 1'b0;
   assign FCMAPUSLEEPNOTREADY = 1'b0;
   assign FCMAPURESULT = regfile_rdata;
   assign FCMAPURESULTVALID = (curr_state == STATE_STORE);
   // ask APU to hold load data because FCM is not ready to process it
   assign FCMAPULOADWAIT = ( (curr_state == STATE_IDLE) &
			     APUFCMLOADDVALID );
   
   // unused output signals
   assign FCMAPUINSTRACK = 1'b0;
   assign FCMAPUDECODEBUSY = 1'b0;
   assign FCMAPUDCDGPRWRITE = 1'b0;
   assign FCMAPUDCDRAEN = 1'b0;
   assign FCMAPUDCDRBEN = 1'b0;
   assign FCMAPUDCDPRIVOP = 1'b0;
   assign FCMAPUDCDFORCEALIGN = 1'b0;
   assign FCMAPUDCDXEROVEN = 1'b0;
   assign FCMAPUDCDXERCAEN = 1'b0;
   assign FCMAPUDCDCREN = 1'b0;
   assign FCMAPUEXECRFIELD = 3'b0;
   assign FCMAPUDCDLOAD = 1'b0;
   assign FCMAPUDCDSTORE = 1'b0;
   assign FCMAPUDCDUPDATE = 1'b0;
   assign FCMAPUDCDLDSTBYTE = 1'b0;
   assign FCMAPUDCDLDSTHW = 1'b0;
   assign FCMAPUDCDLDSTWD = 1'b0;
   assign FCMAPUDCDLDSTDW = 1'b0;
   assign FCMAPUDCDLDSTQW = 1'b0;
   assign FCMAPUDCDTRAPLE = 1'b0;
   assign FCMAPUDCDTRAPBE = 1'b0;
   assign FCMAPUDCDFORCEBESTEERING = 1'b0;
   assign FCMAPUDCDFPUOP = 1'b0;
   assign FCMAPUEXEBLOCKINGMCO = 1'b0;
   assign FCMAPUEXENONBLOCKINGMCO = 1'b0;
   assign FCMAPUXEROV = 1'b0;
   assign FCMAPUXERCA = 1'b0;
   assign FCMAPUCR = 4'b0;
   assign FCMAPUEXCEPTION = 1'b0;
   
   /**************************************************************************/
   /************************** IDCT-interface logic **************************/
   /**************************************************************************/

   /****************************** sequential blocks **************************/   
   // register file for stores
   always @(posedge clock or posedge reset)
     begin
	if (reset) // set all register values to zero
	  begin
	     regfile_store[0]  <= 32'b0; regfile_store[1]  <= 32'b0; 
	     regfile_store[2]  <= 32'b0; regfile_store[3]  <= 32'b0; 
	     regfile_store[4]  <= 32'b0; regfile_store[5]  <= 32'b0; 
	     regfile_store[6]  <= 32'b0; regfile_store[7]  <= 32'b0; 
	     regfile_store[8]  <= 32'b0; regfile_store[9]  <= 32'b0; 
	     regfile_store[10] <= 32'b0; regfile_store[11] <= 32'b0; 
	     regfile_store[12] <= 32'b0; regfile_store[13] <= 32'b0; 
	     regfile_store[14] <= 32'b0; regfile_store[15] <= 32'b0; 
	  end // if (reset)
	else
	  if (regfile_store_we) // write data to specified register when write enabled
	    begin
	       regfile_store[regfile_store_base+4'h1] <= { idctfcmdata[72:79],
							   idctfcmdata[88:95],
							   idctfcmdata[104:111],
							   idctfcmdata[120:127]
							   };
	       regfile_store[regfile_store_base+4'h0] <= { idctfcmdata[8:15],
							   idctfcmdata[24:31],
							   idctfcmdata[40:47],
							   idctfcmdata[56:63]
							   };
	    end
     end

   // pixel row counter
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  pixelrow_counter <= 3'b0;
	else
	  if (pixelrow_counter_we)
	    pixelrow_counter <= pixelrow_counter + 1;
     end

   // state machine for idct interface
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  idct_curr_state <= IDCT_STATE_IDLE;
	else
	  idct_curr_state <= idct_next_state;
     end

   /**************************** combinational blocks *************************/
   // eight to one mux for store regfile base address (row address)
   mux_8_1 regfile_addr_base_mux 
     (
      .outdata(regfile_store_base),
      .indata0(4'b0000),
      .indata1(4'b0010),
      .indata2(4'b0100),
      .indata3(4'b0110),
      .indata4(4'b1000),
      .indata5(4'b1010),
      .indata6(4'b1100),
      .indata7(4'b1110),
      .select(pixelrow_counter)
      );
   
   // state machine for idct interface
   always @(idct_curr_state or FCMAPUDONE or store_or_loadn_reg or
	    ldst_size_reg or pixelrow_counter or idctfcmready)
     begin
	case(idct_curr_state)

	  // wait for quad-word load to be done
	  IDCT_STATE_IDLE:
	    if (~store_or_loadn_reg & FCMAPUDONE &
		(ldst_size_reg == 3) )
	      idct_next_state = IDCT_STATE_SENDDATA;
	    else
	      idct_next_state = IDCT_STATE_IDLE;

	  // send 4 words / 1 row of pixel data to idct module
	  // wait for all 32 words / 8 rows to be sent before
	  // waiting for idct result
	  IDCT_STATE_SENDDATA:
	    if (pixelrow_counter < NUM_PIXELROW)
	      idct_next_state = IDCT_STATE_IDLE;
	    else
	      idct_next_state = IDCT_STATE_WAITIDCT;

	  // wait for idct result
	  IDCT_STATE_WAITIDCT:
	    if (idctfcmready) // 1st output row is available
	      idct_next_state = IDCT_STATE_RECVDATA;
	    else
	      idct_next_state = IDCT_STATE_WAITIDCT;

	  // write all rows of results
	  IDCT_STATE_RECVDATA:
	    if (pixelrow_counter < NUM_PIXELROW)
	      idct_next_state = IDCT_STATE_RECVDATA;
	    else
	      idct_next_state = IDCT_STATE_IDLE;
	  
	  default:
	    idct_next_state = IDCT_STATE_IDLE;
	  
	endcase // case(idct_curr_state)
     end // always @ (idct_curr_state or FCMAPUDONE or store_or_loadn or...

   // internal signal assignments
   assign fcmidctready = (idct_curr_state == IDCT_STATE_SENDDATA);
   // data to idct module from load regfile
   assign fcmidctdata = {regfile[0], 
			 regfile[1],
			 regfile[2],
			 regfile[3]};
   // pad unused bits of idctfcmdata
   assign idctfcmdata[113:116] = idctfcmdata[112] ? 4'b1111 : 4'b0;
   assign idctfcmdata[97:100]  = idctfcmdata[96]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[81:84]   = idctfcmdata[80]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[65:68]   = idctfcmdata[64]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[49:52]   = idctfcmdata[48]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[33:36]   = idctfcmdata[32]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[17:20]   = idctfcmdata[16]  ? 4'b1111 : 4'b0;
   assign idctfcmdata[1:4]     = idctfcmdata[0]   ? 4'b1111 : 4'b0;
   
   // pixelrow counter write enable
   assign pixelrow_counter_we = ( (idct_curr_state == IDCT_STATE_SENDDATA) |
				  idctfcmready);
   assign regfile_store_we = idctfcmready;
   
   /**************************************************************************/
   /******************************* IDCT module ******************************/
   /**************************************************************************/
      
   dct2d_top_final3 idct_0 
     (
      // outputs
      .rdy_out(idctfcmready),
      .out_7( {idctfcmdata[112], idctfcmdata[117:127]} ), 
      .out_6( {idctfcmdata[96],  idctfcmdata[101:111]} ), 
      .out_5( {idctfcmdata[80],  idctfcmdata[85:95]} ), 
      .out_4( {idctfcmdata[64],  idctfcmdata[69:79]} ), 
      .out_3( {idctfcmdata[48],  idctfcmdata[53:63]} ), 
      .out_2( {idctfcmdata[32],  idctfcmdata[37:47]} ), 
      .out_1( {idctfcmdata[16],  idctfcmdata[21:31]} ), 
      .out_0( {idctfcmdata[0],   idctfcmdata[5:15]} ), 
      // inputs
      .rdy_in(fcmidctready), 
      .dct_7( fcmidctdata[116:127] ),
      .dct_6( fcmidctdata[100:111] ),
      .dct_5( fcmidctdata[84:95]   ),
      .dct_4( fcmidctdata[68:79]   ),
      .dct_3( fcmidctdata[52:63]   ),
      .dct_2( fcmidctdata[36:47]   ),
      .dct_1( fcmidctdata[20:31]   ),
      .dct_0( fcmidctdata[4:15]    ),
      .ce_sg(1'b1), // chip enable
      .clk_sg(clock),
      .rst(reset)
      );
   
endmodule // fcm

// decoder for load/store instructions
module decode_ldst
  (
   // outputs
   update,
   size,
   store_or_loadn,
   valid_ldst,
   // inputs
   APUFCMINSTRUCTION
   );

   // output signals
   output       update; // 1=RA is loaded with effective address
   output [0:1] size; // transaction size
   reg [0:1] 	size;
   output 	store_or_loadn; // 1=store, 0=load
   output 	valid_ldst; // if this instruction is a valid FCM load/store

   // input signals
   input [0:31] APUFCMINSTRUCTION;
   
   assign 	update = APUFCMINSTRUCTION[21];
   assign 	store_or_loadn = APUFCMINSTRUCTION[23];
   assign 	valid_ldst = ( (APUFCMINSTRUCTION[0:5]   == 6'b011111) & 
			       (APUFCMINSTRUCTION[26:31] == 6'b001110) );

   always @(APUFCMINSTRUCTION[22] or APUFCMINSTRUCTION[24:25])
     begin
	case({APUFCMINSTRUCTION[22], APUFCMINSTRUCTION[24:25]})

	  3'b100: size = 2'b01; // double-word
	  3'b011: size = 2'b11; // quad-word
	  3'b111: size = 2'b11; // quad-word
	  default: size = 2'b0;
	  
	endcase // case({APUFCMINSTRUCTION[22], APUFCMINSTRUCTION[24:25]})
     end

endmodule // decode

// eight to one multiplexor
module mux_8_1
  (
   // output
   outdata,
   // inputs
   indata0,
   indata1,
   indata2,
   indata3,
   indata4,
   indata5,
   indata6,
   indata7,
   select   
   );

   parameter width = 5;
   
   output [0:(width-1)] outdata;
   reg [0:(width-1)] 	outdata;
   input [0:(width-1)] 	indata0;
   input [0:(width-1)] 	indata1;
   input [0:(width-1)] 	indata2;
   input [0:(width-1)] 	indata3;
   input [0:(width-1)] 	indata4;
   input [0:(width-1)] 	indata5;
   input [0:(width-1)] 	indata6;
   input [0:(width-1)] 	indata7;
   input [0:2] 		select;

   always @(indata0 or indata1 or indata2 or indata3 or
	    indata4 or indata5 or indata6 or indata7 or 
	    select)
     case(select)
       3'b000: outdata = indata0;
       3'b001: outdata = indata1;
       3'b010: outdata = indata2;
       3'b011: outdata = indata3;
       3'b100: outdata = indata4;
       3'b101: outdata = indata5;
       3'b110: outdata = indata6;
       3'b111: outdata = indata7;
     endcase
   
endmodule // mux_8_1
