
# Check to see whether doing behavioral or structural simulation
# based upon the current directory
set current_dir [pwd]
if { [ regexp {simulation/behavioral$} $current_dir ] } {
    echo "Performing a behavioral simulation."
    set BEHAVIORAL_SIM 1
} elseif { [ regexp {simulation/structural$} $current_dir ] } {
    echo "Performing a structural simulation."
    set BEHAVIORAL_SIM 0
} else {
    echo "Current directory is not simulation/behavioral nor simulation/structural"
    error "Testbench supports only behavioral and structural simulation...quitting"
}

# Set PPC_MODEL_BFM to 0 if using SWIFT MODEL of PPC405 (default)
set PPC_MODEL_BFM 0

# echo simulation type
if {$PPC_MODEL_BFM == 1} {
    echo "Using Bus-Functional Model (BFM) of PPC405"
} else {
    echo "Using SWIFT Model of PPC405"
}

# Alter apu_loadstore_init.v file to use the testbench hierarchy
if { [file readable "apu_loadstore_init.v"] } {
    echo
    echo "Modifying apu_loadstore_init.v to use the top-level instance name"
    echo
    exec sh ../../data/init_patch.sh
}

# compile design files
do apu_loadstore.do

# compile testbench models
set testbench_defines "+define+PPC_MODEL_SWIFT"
set testbench_defines "$testbench_defines+BEHAVIORAL_SIM"
vlog $testbench_defines -vlog95compat -incr -work work ../../data/testbench.v

# compile in SWIFT files
vsim -t 1ps -c work.testbench work.apu_loadstore work.glbl

# start wave viewer
do ../../data/wave.do
