/******************************************************************************
 *
 *     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 *     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 *     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 *     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 *     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 *     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 *     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 *     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 *     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 *     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 *     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 *     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *     FOR A PARTICULAR PURPOSE.
 *     
 *     (c) Copyright 2005 Xilinx, Inc.
 *     All rights reserved.
 *
 ******************************************************************************/

/********************************************************************
 * FCM module to test APU-FCM execution of FCM Load and and FCM Store 
 * instructions. This modules contains an interface to the APU and 
 * has a 4-entry register file with 32-bit registers. 
 * 
 * Author: SEG
 * Date: Aug 17, 2004
 * History:
 * 
 *******************************************************************/
`timescale 1 ns / 1 ps

module fcm_loadstore 
  (
   // outputs 
   // to APU
   FCMAPUINSTRACK,
   FCMAPURESULT,
   FCMAPUDONE,
   FCMAPUSLEEPNOTREADY,
   FCMAPUDECODEBUSY,
   FCMAPUDCDGPRWRITE,
   FCMAPUDCDRAEN,
   FCMAPUDCDRBEN,
   FCMAPUDCDPRIVOP,
   FCMAPUDCDFORCEALIGN,
   FCMAPUDCDXEROVEN,
   FCMAPUDCDXERCAEN,
   FCMAPUDCDCREN,
   FCMAPUEXECRFIELD,
   FCMAPUDCDLOAD,
   FCMAPUDCDSTORE,
   FCMAPUDCDUPDATE,
   FCMAPUDCDLDSTBYTE,
   FCMAPUDCDLDSTHW,
   FCMAPUDCDLDSTWD,
   FCMAPUDCDLDSTDW,
   FCMAPUDCDLDSTQW,
   FCMAPUDCDTRAPLE,
   FCMAPUDCDTRAPBE,
   FCMAPUDCDFORCEBESTEERING,
   FCMAPUDCDFPUOP,
   FCMAPUEXEBLOCKINGMCO,
   FCMAPUEXENONBLOCKINGMCO,
   FCMAPULOADWAIT,
   FCMAPURESULTVALID,
   FCMAPUXEROV,
   FCMAPUXERCA,
   FCMAPUCR,
   FCMAPUEXCEPTION,

   // inputs 
   // from APU
   APUFCMINSTRUCTION,
   APUFCMINSTRVALID,
   APUFCMRADATA,
   APUFCMRBDATA,
   APUFCMOPERANDVALID,
   APUFCMFLUSH,
   APUFCMWRITEBACKOK,
   APUFCMLOADDATA,
   APUFCMLOADDVALID,
   APUFCMLOADBYTEEN,
   APUFCMENDIAN,
   APUFCMXERCA,
   APUFCMDECODED,
   APUFCMDECUDI,
   APUFCMDECUDIVALID,

   // clock and reset
   clock,
   reset
   
   );

   // state mnemonics
   parameter // load/store states
	     STATE_IDLE  = 2'b00, // idle state
	     STATE_LOAD  = 2'b01, // valid load instruction detected
	     STATE_STORE = 2'b10; // valid store instruction detected
	     
   // port declarations
   // outputs
   // to APU
   output        FCMAPUINSTRACK;
   output [0:31] FCMAPURESULT;
   output 	 FCMAPUDONE;
   output 	 FCMAPUSLEEPNOTREADY;
   output 	 FCMAPUDECODEBUSY;
   output 	 FCMAPUDCDGPRWRITE;
   output 	 FCMAPUDCDRAEN;
   output 	 FCMAPUDCDRBEN;
   output 	 FCMAPUDCDPRIVOP;
   output 	 FCMAPUDCDFORCEALIGN;
   output 	 FCMAPUDCDXEROVEN;
   output 	 FCMAPUDCDXERCAEN;
   output 	 FCMAPUDCDCREN;
   output [0:2]  FCMAPUEXECRFIELD;
   output 	 FCMAPUDCDLOAD;
   output 	 FCMAPUDCDSTORE;
   output 	 FCMAPUDCDUPDATE;
   output 	 FCMAPUDCDLDSTBYTE;
   output 	 FCMAPUDCDLDSTHW;
   output 	 FCMAPUDCDLDSTWD;
   output 	 FCMAPUDCDLDSTDW;
   output 	 FCMAPUDCDLDSTQW;
   output 	 FCMAPUDCDTRAPLE;
   output 	 FCMAPUDCDTRAPBE;
   output 	 FCMAPUDCDFORCEBESTEERING;
   output 	 FCMAPUDCDFPUOP;
   output 	 FCMAPUEXEBLOCKINGMCO;
   output 	 FCMAPUEXENONBLOCKINGMCO;
   output 	 FCMAPULOADWAIT;
   output 	 FCMAPURESULTVALID;
   output 	 FCMAPUXEROV;
   output 	 FCMAPUXERCA;
   output [0:3]  FCMAPUCR;
   output 	 FCMAPUEXCEPTION;
	
   // inputs
   // to APU
   input [0:31]  APUFCMINSTRUCTION;
   input 	 APUFCMINSTRVALID;
   input [0:31]	 APUFCMRADATA;
   input [0:31]	 APUFCMRBDATA;
   input 	 APUFCMOPERANDVALID;
   input 	 APUFCMFLUSH;
   input 	 APUFCMWRITEBACKOK;
   input [0:31]  APUFCMLOADDATA;
   input 	 APUFCMLOADDVALID;
   input [0:3] 	 APUFCMLOADBYTEEN;
   input 	 APUFCMENDIAN;
   input 	 APUFCMXERCA;
   input 	 APUFCMDECODED;
   input [0:2] 	 APUFCMDECUDI;
   input 	 APUFCMDECUDIVALID;
   // clock and reset
   input 	 clock;
   input 	 reset;
   
   // internal signals
   // PPC instruction-related
   reg [0:5] 	 reg_RT; // target register of PPC instruction
   reg [0:5] 	 reg_RA; // base register of PPC instruction
   reg [0:5] 	 reg_RB; // offset register of PPC instruction
   wire 	 instrreg_we; // write enable
   // loads and stores
   wire 	 store_or_loadn; // 1=store, 0=load
   reg 		 store_or_loadn_reg; // stores the store_or_loadn signal
   wire 	 ldst_update; // 1=update RA
   wire [0:1] 	 ldst_size; // number of words to transfer
   reg [0:1] 	 ldst_size_reg; // stores number of words to transfer
   reg [0:1] 	 ldst_size_counter; // counter for number of words
   wire 	 ldst_size_counter_we; // write enable   
   wire 	 ldst_valid; // 1=valid load/store instruction   
   // register file for FCM loads
   reg [0:31] 	 regfile [0:3]; // 32-bit entry register file with 4 registers
   wire 	 regfile_we; // write enable
   wire [0:1] 	 regfile_waddr; // write address
   wire [0:31] 	 regfile_wdata; // write data
   reg [0:1] 	 regfile_raddr; // register to store read address
   wire 	 regfile_raddr_we; // write enable 	 
   wire [0:31] 	 regfile_rdata; // read data
   // state registers
   reg [0:1] 	 curr_state; // current state
   reg [0:1] 	 next_state; // next state

   // register file for FCM stores
   reg [0:31] 	 regfile_store[0:15]; // 32-bit entry register file with 16 registers
   wire 	 regfile_store_we; // write enable
   wire [0:3] 	 regfile_store_base; // block (1 block == 4 words) addr
   
   /****************************** sequential blocks **************************/
   // PPC instruction-related registers
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  begin
	     reg_RT <= 5'b0;
	     reg_RA <= 5'b0;
	     reg_RB <= 5'b0;
	     store_or_loadn_reg <= 1'b0;
	     ldst_size_reg <= 2'b0;
	  end
	else
	  if (instrreg_we) // capture instruction information
	    begin
	       reg_RT <= APUFCMINSTRUCTION[6:10];
	       reg_RA <= APUFCMINSTRUCTION[11:15];
	       reg_RB <= APUFCMINSTRUCTION[16:20];
	       store_or_loadn_reg <= store_or_loadn;
	       ldst_size_reg <= ldst_size;
	    end
     end

   // read address of register file
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  regfile_raddr <= 5'h0;
	else
	  if (instrreg_we)
	    regfile_raddr <= APUFCMINSTRUCTION[6:10];
	  else
	    if (regfile_raddr_we)
	      regfile_raddr <= regfile_raddr + 1;
     end // always @ (posedge clock or posedge reset)
   
   // load / store counter for number of words
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  ldst_size_counter <= 2'b0;
	else
	  if (instrreg_we)
	    ldst_size_counter <= 2'b0;
	  else
	    if (ldst_size_counter_we)
	      ldst_size_counter <= ldst_size_counter + 1;
     end // always @ (posedge clock or posedge reset)

   // register file
   always @(posedge clock or posedge reset)
     begin
	if (reset) // set all register values to zero
	  begin
	     regfile[0]  <= 32'b0; regfile[1]  <= 32'b0; 
	     regfile[2]  <= 32'b0; regfile[3]  <= 32'b0; 
	  end // if (reset)
	else
	  if (regfile_we) // write data to specified register when write enabled
	    regfile[regfile_waddr] <= regfile_wdata;
     end

   // state machine
   always @(posedge clock or posedge reset)
     begin
	if (reset)
	  begin
	     curr_state <= STATE_IDLE;
	  end
	else
	  curr_state <= next_state;
     end

   /**************************** combinational blocks *************************/
   // decoder
   decode_ldst decode_ldst_0 
     ( // outputs
       .update(ldst_update),
       .size(ldst_size),
       .store_or_loadn(store_or_loadn),
       .valid_ldst(ldst_valid),
       // inputs
       .APUFCMINSTRUCTION(APUFCMINSTRUCTION)
       );
   
   // state machine logic
   always @(curr_state or store_or_loadn or store_or_loadn_reg or 
	    ldst_size_counter or ldst_size_reg or
	    APUFCMINSTRVALID or 
	    APUFCMLOADDVALID or
	    APUFCMDECODED)
     begin
	case (curr_state)

	  // wait for valid instruction
	  STATE_IDLE:
	    if (APUFCMINSTRVALID &
		APUFCMDECODED) // valid instruction from APU
	      if (store_or_loadn) // store instruction
		next_state = STATE_STORE;
	      else // load instruction
		if (APUFCMLOADDVALID) // load data arrived at the same time
		  if (ldst_size_counter < ldst_size_reg)
		    next_state = STATE_LOAD;
		  else
		    next_state = STATE_IDLE;
		else
		  next_state = STATE_LOAD;
	    else
	      next_state = STATE_IDLE;

	  // seen a valid load instruction, wait for valid data
	  STATE_LOAD:
	    // keep track of how many words to access
	    if (ldst_size_counter < ldst_size_reg)
	      next_state = STATE_LOAD;
	    else
	      if (APUFCMLOADDVALID)
		next_state = STATE_IDLE;
	      else
		next_state = STATE_LOAD;
	  
	  // seen a valid store instruction, output data
	  STATE_STORE:
	    // keep track of how many words to access
	    if (ldst_size_counter < ldst_size_reg)
	      next_state = STATE_STORE;
	    else
	      next_state = STATE_IDLE;
	  
	  default:
	    next_state = STATE_IDLE;
	  
	endcase // case(curr_state)
     end // always @ (APUFCMINSTRUCTION or APUFCMINSTRVALID or ...

   // internal signal assignments
   assign instrreg_we = (curr_state == STATE_IDLE);
   // data might arrive at the same time as the instruction
   assign regfile_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID) |
			 ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID) );
   // reg_RT is the target register value when data comes after the instruction
   assign regfile_waddr = (reg_RT + ldst_size_counter);
   // write data from memory to register file
   assign regfile_wdata = APUFCMLOADDATA;
   // address of register file to read data from for a store operation
   assign regfile_raddr_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID ) |
			       (curr_state == STATE_STORE) |
			       ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID ) );
   // data read for a store operation
   assign regfile_rdata = regfile[regfile_raddr];
   // update counter for number of data words transferred
   assign ldst_size_counter_we = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID ) |
				   (curr_state == STATE_STORE) |
				   ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID ) );
   
   // output assignments
   assign FCMAPUDONE = ( ( (curr_state == STATE_LOAD) & APUFCMLOADDVALID &
			   (ldst_size_counter == ldst_size_reg) ) |
			 ( (curr_state == STATE_STORE) & 
			   (ldst_size_counter == ldst_size_reg) ) |
			 ( (curr_state == STATE_IDLE) & APUFCMLOADDVALID &
			   (ldst_size_counter == ldst_size_reg) ) ) ? 1'b1 : 1'b0;
   assign FCMAPUSLEEPNOTREADY = 1'b0;
   assign FCMAPURESULT = regfile_rdata;
   assign FCMAPURESULTVALID = (curr_state == STATE_STORE);
   // ask APU to hold load data because FCM is not ready to process it
   assign FCMAPULOADWAIT = ( (curr_state == STATE_IDLE) &
			     APUFCMLOADDVALID );
   
   // unused output signals
   assign FCMAPUINSTRACK = 1'b0;
   assign FCMAPUDECODEBUSY = 1'b0;
   assign FCMAPUDCDGPRWRITE = 1'b0;
   assign FCMAPUDCDRAEN = 1'b0;
   assign FCMAPUDCDRBEN = 1'b0;
   assign FCMAPUDCDPRIVOP = 1'b0;
   assign FCMAPUDCDFORCEALIGN = 1'b0;
   assign FCMAPUDCDXEROVEN = 1'b0;
   assign FCMAPUDCDXERCAEN = 1'b0;
   assign FCMAPUDCDCREN = 1'b0;
   assign FCMAPUEXECRFIELD = 3'b0;
   assign FCMAPUDCDLOAD = 1'b0;
   assign FCMAPUDCDSTORE = 1'b0;
   assign FCMAPUDCDUPDATE = 1'b0;
   assign FCMAPUDCDLDSTBYTE = 1'b0;
   assign FCMAPUDCDLDSTHW = 1'b0;
   assign FCMAPUDCDLDSTWD = 1'b0;
   assign FCMAPUDCDLDSTDW = 1'b0;
   assign FCMAPUDCDLDSTQW = 1'b0;
   assign FCMAPUDCDTRAPLE = 1'b0;
   assign FCMAPUDCDTRAPBE = 1'b0;
   assign FCMAPUDCDFORCEBESTEERING = 1'b0;
   assign FCMAPUDCDFPUOP = 1'b0;
   assign FCMAPUEXEBLOCKINGMCO = 1'b0;
   assign FCMAPUEXENONBLOCKINGMCO = 1'b0;
   assign FCMAPUXEROV = 1'b0;
   assign FCMAPUXERCA = 1'b0;
   assign FCMAPUCR = 4'b0;
   assign FCMAPUEXCEPTION = 1'b0;
      
endmodule // fcm_loadstore

// decoder for load/store instructions
module decode_ldst
  (
   // outputs
   update,
   size,
   store_or_loadn,
   valid_ldst,
   // inputs
   APUFCMINSTRUCTION
   );

   // output signals
   output       update; // 1=RA is loaded with effective address
   output [0:1] size; // transaction size
   reg [0:1] 	size;
   output 	store_or_loadn; // 1=store, 0=load
   output 	valid_ldst; // if this instruction is a valid FCM load/store

   // input signals
   input [0:31] APUFCMINSTRUCTION;
   
   assign 	update = APUFCMINSTRUCTION[21];
   assign 	store_or_loadn = APUFCMINSTRUCTION[23];
   assign 	valid_ldst = ( (APUFCMINSTRUCTION[0:5]   == 6'b011111) & 
			       (APUFCMINSTRUCTION[26:31] == 6'b001110) );

   always @(APUFCMINSTRUCTION[22] or APUFCMINSTRUCTION[24:25])
     begin
	case({APUFCMINSTRUCTION[22], APUFCMINSTRUCTION[24:25]})

	  3'b100: size = 2'b01; // double-word
	  3'b011: size = 2'b11; // quad-word
	  3'b111: size = 2'b11; // quad-word
	  default: size = 2'b0;
	  
	endcase // case({APUFCMINSTRUCTION[22], APUFCMINSTRUCTION[24:25]})
     end

endmodule // decode_ldst