/******************************************************************************
 *
 *     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 *     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 *     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 *     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 *     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 *     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 *     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 *     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 *     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 *     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 *     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 *     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *     FOR A PARTICULAR PURPOSE.
 *     
 *     (c) Copyright 2005 Xilinx, Inc.
 *     All rights reserved.
 *
 ******************************************************************************/

#include <string.h>
#include "xparameters.h"
#include "xbasic_types.h"
#include "xcache_l.h"
#include "xgpio_l.h"
#include "xtime_l.h"
#include "xtft.h"
#include "xtft_l.h"
#include "xpseudo_asm.h"
#include "bootload_basicgraphics.h"
#include "xromlcd.h"

/* Pushbutton base address */
#define GPIO_0_BASEADDR XPAR_OPB_GPIO_0_BASEADDR
/* TFT display base address */
#define TFT_DEVICE_ID   XPAR_PLB_TFT_CNTLR_REF_0_DEVICE_ID
#define DISP_BASE       (62*1024*1024)
#define SWITCH_INTERVAL 0x59682F00
/* keep track of which IDCT implementation the system is running */
static short is_hw_idct = 0; /* hardware idct by default */

/************************ Software IDCT *************************/

/*
 * idct.c
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * Portions of this code are from the MPEG software simulation group
 * idct implementation. This code will be replaced with a new
 * implementation soon.
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**********************************************************/
/* inverse two dimensional DCT, Chen-Wang algorithm */
/* (cf. IEEE ASSP-32, pp. 803-816, Aug. 1984) */
/* 32-bit integer arithmetic (8 bit coefficients) */
/* 11 mults, 29 adds per DCT */
/* sE, 18.8.91 */
/**********************************************************/
/* coefficients extended to 12 bit for IEEE1180-1990 */
/* compliance sE, 2.1.94 */
/**********************************************************/

/* this code assumes >> to be a two's-complement arithmetic */
/* right shift: (-2)>>1 == -1 , (-3)>>1 == -2 */

#define uint8_t  unsigned char
#define int16_t  short int
#define uint32_t unsigned int

#ifdef __DCC__
#define inline
#endif

#define W1 2841 /* 2048*sqrt (2)*cos (1*pi/16) */
#define W2 2676 /* 2048*sqrt (2)*cos (2*pi/16) */
#define W3 2408 /* 2048*sqrt (2)*cos (3*pi/16) */
#define W5 1609 /* 2048*sqrt (2)*cos (5*pi/16) */
#define W6 1108 /* 2048*sqrt (2)*cos (6*pi/16) */
#define W7 565  /* 2048*sqrt (2)*cos (7*pi/16) */

//static uint8_t clip_lut[1024];
//#define CLIP(i) ((clip_lut+384)[ (i)])

/* row (horizontal) IDCT
 *
 * 7 pi 1
 * dst[k] = sum c[l] * src[l] * cos ( -- * ( k + - ) * l )
 * l=0 8 2
 *
 * where: c[0] = 128
 * c[1..7] = 128*sqrt (2)
 */

static void inline idct_row (int16_t * block, int16_t * dblock)
{
    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

    x1 = block[4] << 11;
    x2 = block[6];
    x3 = block[2];
    x4 = block[1];
    x5 = block[7];
    x6 = block[5];
    x7 = block[3];

    /* shortcut */
    if (! (x1 | x2 | x3 | x4 | x5 | x6 | x7 )) {
	dblock[0] = dblock[1] = dblock[2] = dblock[3] = dblock[4] =
	    dblock[5] = dblock[6] = dblock[7] = block[0]<<3;
	return;
    }

    x0 = (block[0] << 11) + 128; /* for proper rounding in the fourth stage */

    /* first stage */
    x8 = W7 * (x4 + x5);
    x4 = x8 + (W1 - W7) * x4;
    x5 = x8 - (W1 + W7) * x5;
    x8 = W3 * (x6 + x7);
    x6 = x8 - (W3 - W5) * x6;
    x7 = x8 - (W3 + W5) * x7;
 
    /* second stage */
    x8 = x0 + x1;
    x0 -= x1;
    x1 = W6 * (x3 + x2);
    x2 = x1 - (W2 + W6) * x2;
    x3 = x1 + (W2 - W6) * x3;
    x1 = x4 + x6;
    x4 -= x6;
    x6 = x5 + x7;
    x5 -= x7;
 
    /* third stage */
    x7 = x8 + x3;
    x8 -= x3;
    x3 = x0 + x2;
    x0 -= x2;
    x2 = (181 * (x4 + x5) + 128) >> 8;
    x4 = (181 * (x4 - x5) + 128) >> 8;
 
    /* fourth stage */
    dblock[0] = (x7 + x1) >> 8;
    dblock[1] = (x3 + x2) >> 8;
    dblock[2] = (x0 + x4) >> 8;
    dblock[3] = (x8 + x6) >> 8;
    dblock[4] = (x8 - x6) >> 8;
    dblock[5] = (x0 - x4) >> 8;
    dblock[6] = (x3 - x2) >> 8;
    dblock[7] = (x7 - x1) >> 8;
}

/* column (vertical) IDCT
 *
 * 7 pi 1
 * dst[8*k] = sum c[l] * src[8*l] * cos ( -- * ( k + - ) * l )
 * l=0 8 2
 *
 * where: c[0] = 1/1024
 * c[1..7] = (1/1024)*sqrt (2)
 */

static void inline idct_col (int16_t *block, unsigned char *dblock)
{
    int x0, x1, x2, x3, x4, x5, x6, x7, x8;

    /* shortcut */
    x1 = block [8*4] << 8;
    x2 = block [8*6];
    x3 = block [8*2];
    x4 = block [8*1];
    x5 = block [8*7];
    x6 = block [8*5];
    x7 = block [8*3];

#if 0
    if (! (x1 | x2 | x3 | x4 | x5 | x6 | x7 )) {
	block[8*0] = block[8*1] = block[8*2] = block[8*3] = block[8*4] =
	    block[8*5] = block[8*6] = block[8*7] = (block[8*0] + 32) >> 6;
	return;
    }
#endif

    x0 = (block[8*0] << 8) + 8192;

    /* first stage */
    x8 = W7 * (x4 + x5) + 4;
    x4 = (x8 + (W1 - W7) * x4) >> 3;
    x5 = (x8 - (W1 + W7) * x5) >> 3;
    x8 = W3 * (x6 + x7) + 4;
    x6 = (x8 - (W3 - W5) * x6) >> 3;
    x7 = (x8 - (W3 + W5) * x7) >> 3;
 
    /* second stage */
    x8 = x0 + x1;
    x0 -= x1;
    x1 = W6 * (x3 + x2) + 4;
    x2 = (x1 - (W2 + W6) * x2) >> 3;
    x3 = (x1 + (W2 - W6) * x3) >> 3;
    x1 = x4 + x6;
    x4 -= x6;
    x6 = x5 + x7;
    x5 -= x7;
 
    /* third stage */
    x7 = x8 + x3;
    x8 -= x3;
    x3 = x0 + x2;
    x0 -= x2;
    x2 = (181 * (x4 + x5) + 128) >> 8;
    x4 = (181 * (x4 - x5) + 128) >> 8;
 
    /* fourth stage */
    dblock[8*0] = (unsigned char) ((x7 + x1) >> 14);
    dblock[8*1] = (unsigned char) ((x3 + x2) >> 14);
    dblock[8*2] = (unsigned char) ((x0 + x4) >> 14);
    dblock[8*3] = (unsigned char) ((x8 + x6) >> 14);
    dblock[8*4] = (unsigned char) ((x8 - x6) >> 14);
    dblock[8*5] = (unsigned char) ((x0 - x4) >> 14);
    dblock[8*6] = (unsigned char) ((x3 - x2) >> 14);
    dblock[8*7] = (unsigned char) ((x7 - x1) >> 14);
}

static void inline mpeg2_idct_c_sw (int16_t * block, unsigned char * dblock,
				 int disp)
{
    int i;
    int16_t tblock[64];
    unsigned char tdblock[64];

    for (i = 0; i < 8; i++)
	idct_row (block + 8 * i, tblock + 8 * i);

    for (i = 0; i < 8; i++)
	idct_col (tblock + i, tdblock + i);
    
    for (i = 0; i < 8; i++)
      memcpy(dblock + i*disp, tdblock + i*8, 8);
}


/************************ Hardware IDCT *************************/

#define dcba(adr)    __asm__ __volatile__(\
                           "dcba 0,%0\n"\
                           : : "r" (adr)\
                         )

#define lqfcmx(rn, base, adr)    __asm__ __volatile__(\
                           "lqfcmx " #rn ",%0,%1\n"\
                           : : "b" (base), "r" (adr)\
                         )

#define stdfcmx(rn, base, adr)    __asm__ __volatile__(\
                                   "stdfcmx " #rn ",%0,%1\n"\
                                   : : "b" (base), "r" (adr)\
                                 )

static void mpeg2_idct_c_hw (int16_t * block, uint8_t * resultblock)
{
  // load idct input data from memory
  lqfcmx(0, block, 0*16);
  lqfcmx(0, block, 1*16);
  lqfcmx(0, block, 2*16);
  lqfcmx(0, block, 3*16);
  lqfcmx(0, block, 4*16);
  lqfcmx(0, block, 5*16);
  lqfcmx(0, block, 6*16);
  lqfcmx(0, block, 7*16);
    
  // store result into memory
  stdfcmx(0, resultblock, 0*32);
  stdfcmx(2, resultblock, 1*32);
  stdfcmx(4, resultblock, 2*32);
  stdfcmx(6, resultblock, 3*32);
  stdfcmx(8, resultblock, 4*32);
  stdfcmx(10, resultblock, 5*32);
  stdfcmx(12, resultblock, 6*32);
  stdfcmx(14, resultblock, 7*32);
}

//uint8_t block[4*64] __attribute__ ((aligned(32))) = {0};
uint8_t block_buf[4096] __attribute__ ((aligned(4096))) = {0};
uint8_t *block = &block_buf[4*480];

/* draw sidebar with performance graph */
void draw_sidebar(XTft Tft) {
  unsigned int bgCol = 0x007F9F90;
  /* title */
  XTft_DrawSolidBox(&Tft, 450, 0, 639, 479, bgCol);
  XTft_SetPos(&Tft, 531, 20);
  XTft_WriteColorString(&Tft, "Enhanced", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 524, 35);
  XTft_WriteColorString(&Tft, "Processing", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 526, 50);
  XTft_WriteColorString(&Tft, "using the", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 526, 70);
  XTft_WriteColorString(&Tft, "Auxiliary", 0x0000FFFF, bgCol);
  XTft_SetPos(&Tft, 527, 85);
  XTft_WriteColorString(&Tft, "Processor", 0x0000FFFF, bgCol);
  XTft_SetPos(&Tft, 525, 100);
  XTft_WriteColorString(&Tft, "Unit (APU)", 0x0000FFFF, bgCol);
  /* white box */
  XTft_DrawSolidBox(&Tft, 500, 140, 619, 399, 0x00FFFFFF);
  XTft_SetPos(&Tft, 540, 145);
  XTft_WriteColorString(&Tft, "IDCT", 0x00000090, 0x00FFFFFF);
  XTft_DrawLine(&Tft, 540, 156, 570, 156, 0x00000090); // underline
  XTft_SetPos(&Tft, 515, 160);
  XTft_WriteColorString(&Tft, "Performance", 0x00000090, 0x00FFFFFF);
  XTft_DrawLine(&Tft, 515, 171, 601, 171, 0x00000090); // underline

  /* draw scale */
  XTft_DrawSolidBox(&Tft, 580, 192, 582, 341, 0x00000000);
  XTft_DrawLine(&Tft, 578, 192, 585, 192, 0x00000000);
  XTft_DrawLine(&Tft, 578, 267, 585, 267, 0x00000000);
  XTft_DrawLine(&Tft, 532, 342, 585, 342, 0x00000000);
  /* graph - idct without apu acceleration */
  XTft_DrawSolidBox(&Tft, 535, 315, 573, 341, 0x00FF0000); // red

  /* current system - Without APU or With APU */
  XTft_DrawSolidBox(&Tft, 505, 360, 614, 389, 0x00CFCFCF); // grey
  XTft_SetPos(&Tft, 519, 371);
  XTft_WriteColorString(&Tft, "Without APU", 0x00000000, 0x00CFCFCF);

  /* footer */
  XTft_SetPos(&Tft, 520, 435);
  XTft_WriteColorString(&Tft, "Virtex-4 FX", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 535, 450);
  XTft_WriteColorString(&Tft, "Xilinx", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 536, 450);
  XTft_WriteColorString(&Tft, "Xilinx", 0x00FFFFFF, bgCol);
  XTft_SetPos(&Tft, 536, 450);
  XTft_WriteColorString(&Tft, "Xilinx", 0x00FFFFFF, bgCol);
}

/* switch from software to hardware IDCT and vice versa */
void switch_idct(XTft Tft) {
  if (is_hw_idct) { /* do software idct */
    is_hw_idct = 0;
    /* graph - idct without apu acceleration */
    XTft_DrawSolidBox(&Tft, 535, 192, 573, 314, 0x00FFFFFF); // white
    XTft_DrawSolidBox(&Tft, 535, 315, 573, 341, 0x00FF0000); // red
    XTft_DrawSolidBox(&Tft, 505, 360, 614, 389, 0x00CFCFCF); // grey
    XTft_SetPos(&Tft, 519, 371);
    XTft_WriteColorString(&Tft, "Without APU", 0x00000000, 0x00CFCFCF);
  }
  else { /* do hardware idct */
    is_hw_idct = 1;
    /* graph - idct with apu acceleration */
    XTft_DrawSolidBox(&Tft, 535, 192, 573, 341, 0x0000FF00); // green
    XTft_DrawSolidBox(&Tft, 505, 360, 614, 389, 0x00F0E190); // gold
    XTft_SetPos(&Tft, 530, 371);
    XTft_WriteColorString(&Tft, "With APU", 0x00000000, 0x00F0E190); // gold
  }
}

/************************ Main routine **********************/
int main()
{
  XTft Tft;
  Xboolean button_pushed;
  XTime start, end;
  unsigned int p, r, c, i, k;
  unsigned char *src, *dst;
  int dsta, blocka;

  /* enable caches */
  XCache_EnableICache(0x80000001);
  XCache_EnableDCache(0x80000001);
  /* enable APU */
  mtmsr(XREG_MSR_APU_AVAILABLE);
  /* set GPIO to be able to read buttons */
  XGpio_mSetDataDirection(GPIO_0_BASEADDR, 1, 0x1F);
  /* initialize TFT display */
  XTft_Initialize(&Tft, TFT_DEVICE_ID);

  /* write white */
  memset((void *) DISP_BASE, -1, 2*1024*1024);
  /* draw performance graph */
  draw_sidebar(Tft);
  /* start timing */
  XTime_GetTime(&start);

  /* print directions to LCD screen */
  XromLCDOn();
  XromLCDInit();
  XromLCDPrintString ("  APU-Enhanced  ","   Processing   ");

  while(1)
  {
    src = (char *) 0;
    for (p=0; p < 30; p++)
    {
      dst = (char *) DISP_BASE;
      /* poll push-buttons 
	 if either one has been pressed, check flag and 
	 switch to hardware or software idct */
      button_pushed = (XGpio_mGetDataReg(GPIO_0_BASEADDR, 1) & 0x00003E00);
      XTime_GetTime(&end);
      if ( button_pushed || /* pushbutton was pressed */
 	   ((end - start) > SWITCH_INTERVAL) ) /* check time elapsed since last switch */
      {
	start = end;
	switch_idct(Tft);
      }
      for (r=0; r < 60; r++) /* for 60 rows of 480x480 display */
      {
	/* the idct operations */
	if (is_hw_idct)
	{ /* do hardware idct */
	  for (c=0; c < 15*4; c++) /* for 60 columns of 480x480 TFT display */
	  {
	    for (k=0; k < 4; k++) /* 4 IDCTs on DCT values */
	    {
	      mpeg2_idct_c_hw((int16_t *)src, block + k*8);
	      src += 64*2;
	    }
	    for (i=0; i < 8; i++) /* copy each row of IDCT output to TFT memory */
	    {
	      dsta = (int)(dst+i*4096); /* start of display memory */
	      blocka = (int)(block+i*32); /* start of IDCT results */
	      dcba(dsta); /* remove unnecessary cacheline read from a write operation */
	      memcpy(dst+i*4096, block+i*32, 32);
	    }
	    dst += 32;
	  } /* end of 60 columns of 480x480 TFT */
	} /* end of hardware idct */
	else
	{ /* do software idct */
	  for (c=0; c < 60*4; c++)
	  {
	    mpeg2_idct_c_sw((int16_t *)src, dst, 4096);
	    dst += 8;
	    src += 64*2;
	  } /* end of software idct */
	} /* end of switching between hw and sw */
	dst += (8*1024-480)*4; /* advance 8 rows */
      } /* end of 60 rows of 480x480 TFT */
    } /* end of 30 frames */
  }

  return 0;
}

