/******************************************************************************
 *
 *     XILINX IS PROVIDING THIS DESIGN, CODE, OR INFORMATION "AS IS"
 *     SOLELY FOR USE IN DEVELOPING PROGRAMS AND SOLUTIONS FOR
 *     XILINX DEVICES.  BY PROVIDING THIS DESIGN, CODE, OR INFORMATION
 *     AS ONE POSSIBLE IMPLEMENTATION OF THIS FEATURE, APPLICATION
 *     OR STANDARD, XILINX IS MAKING NO REPRESENTATION THAT THIS
 *     IMPLEMENTATION IS FREE FROM ANY CLAIMS OF INFRINGEMENT,
 *     AND YOU ARE RESPONSIBLE FOR OBTAINING ANY RIGHTS YOU MAY REQUIRE
 *     FOR YOUR IMPLEMENTATION.  XILINX EXPRESSLY DISCLAIMS ANY
 *     WARRANTY WHATSOEVER WITH RESPECT TO THE ADEQUACY OF THE
 *     IMPLEMENTATION, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OR
 *     REPRESENTATIONS THAT THIS IMPLEMENTATION IS FREE FROM CLAIMS OF
 *     INFRINGEMENT, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *     FOR A PARTICULAR PURPOSE.
 *     
 *     (c) Copyright 2005 Xilinx, Inc.
 *     All rights reserved.
 *
 ******************************************************************************/

#include "xparameters.h"
#include "xcache_l.h"
#include "xtime_l.h"
#include "xpseudo_asm.h"

#define uint8_t  unsigned char
#define int16_t  short int
#define uint32_t unsigned int

#define lqfcmx(rn, base, adr)    __asm__ __volatile__(\
                           "lqfcmx " #rn ",%0,%1\n"\
                           : : "b" (base), "r" (adr)\
                         )

#define stdfcmx(rn, base, adr)    __asm__ __volatile__(\
                                   "stdfcmx " #rn ",%0,%1\n"\
                                   : : "b" (base), "r" (adr)\
                                 )

#define RUNS    1

volatile int16_t __attribute__ ((aligned (16))) block[] = {214,49,-3,20,-10,-1,1,-6,34,-25,11,13,5,-3,15,-6,-6,-4,8,-9,3,-3,5,10,8,-10,4,4,-15,10,6,6,-12,5,-1,-2,-15,9,-5,-1,5,9,-8,3,4,-7,-14,2,2,-2,3,-1,1,3,-3,-4,-1,1,0,2,3,-2,-4,-2};
volatile uint8_t __attribute__ ((aligned (8))) resultblock[] = {0};

void mpeg2_idct (int16_t *block, int16_t *resultblock)
{
  unsigned int i;

  // load idct input data from memory
  for (i=0; i<8; i++) {
    lqfcmx(0, block, i*16);
  }

  // store result into memory
  stdfcmx(0, resultblock, 0);
  stdfcmx(2, resultblock, 8);
  stdfcmx(4, resultblock, 16);
  stdfcmx(6, resultblock, 24);
  stdfcmx(8, resultblock, 32);
  stdfcmx(10, resultblock, 40);
  stdfcmx(12, resultblock, 48);
  stdfcmx(14, resultblock, 56);
}

int main(void)
{
  int i;
  XTime start, end, diff, idcts_per_sec;

  // initialize caches
  XCache_EnableDCache(0x80000001);
  XCache_EnableICache(0x80000001);
  // initialize APU
  mtmsr(XREG_MSR_APU_AVAILABLE);

  XTime_GetTime(&start);
  // run IDCT
  for (i = 0; i < RUNS; i++) {
    mpeg2_idct((int16_t *)block, (int16_t *)resultblock);
  }
  XTime_GetTime(&end);
  
  // verify results in simulation
  for (i=0; i<16; i++)
    mtgpr(14,((unsigned int *)resultblock)[i]);

  diff = end - start;
  idcts_per_sec = (1ULL * RUNS * XPAR_CPU_PPC405_CORE_CLOCK_FREQ_HZ) / diff;

  return (int) idcts_per_sec;
}
